from argparse import ArgumentParser

from dodo_commands import Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.choice_picker import ChoicePicker

docker = plumbum.cmd.docker


def _args():
    parser = ArgumentParser()
    args = Dodo.parse_args(parser)
    return args


def _containers():
    result = []
    for line in docker("ps", "--format",
                       "{{.ID}} {{.Names}} {{.Image}}").split('\n'):
        if line:
            cid, name, image = line.split()
            result.append(dict(name=name, cid=cid, image=image))
    return result


if Dodo.is_main(__name__):
    args = _args()

    class Picker(ChoicePicker):
        def print_choices(self, choices):
            for idx, container in enumerate(choices):
                print("%d - %s" % (idx + 1, container['name']))

        def question(self):
            return "Select a container: "

    picker = Picker(_containers())
    picker.pick()

    for container in picker.get_choices():
        Dodo.run(['docker', 'kill', container['cid']], )



"""Tests for config.py."""

import os

from dodo_commands.dependencies import yaml_round_trip_dump

import pytest

from ..config import load_config
from ..config_io import ConfigIO


class TestConfigIO:  # noqa
    @pytest.fixture
    def simple_config(self, tmpdir):
        foo_dir = os.path.join(str(tmpdir), 'foo')
        return {
            'LAYERS': ['mylayer.yml'],
            'ROOT': {
                'command_path': [os.path.join(foo_dir, 'bar')],
                'foo': {
                    'bar': 'foobar',
                    'one': 'two'
                },
                'one': 1
            }
        }

    @pytest.fixture
    def layer(self, tmpdir):
        foo_dir = os.path.join(str(tmpdir), 'foo')
        return {
            'ROOT': {
                'command_path': [
                    os.path.join(foo_dir, 'foobar'),
                ],
                'foo': {
                    'bar': 'barfoo',
                    'three': 4
                },
                'one': 1.1
            }
        }

    @pytest.fixture
    def create_config(self, tmpdir, simple_config):
        """Create a config file in the tmpdir."""
        config_io = ConfigIO(str(tmpdir))  # noqa
        config_io.save(simple_config)

    @pytest.fixture
    def create_layer(self, tmpdir, layer):
        """Create a config layer file in the tmpdir."""
        layer_filename = os.path.join(str(tmpdir), "mylayer.yml")
        with open(layer_filename, "w") as f:
            f.write(yaml_round_trip_dump(layer))

    def test_save(self, tmpdir, simple_config):  # noqa
        config_io = ConfigIO(str(tmpdir))  # noqa
        config_io.save(simple_config)
        config = config_io.load()
        assert simple_config == config

    @pytest.mark.usefixtures("create_config", "create_layer")
    def test_layers(self, tmpdir):  # noqa
        foo_dir = os.path.join(str(tmpdir), 'foo')
        config_io = ConfigIO(str(tmpdir))  # noqa
        full_config = load_config(['config.yaml', 'mylayer.yml'],
                                  config_io=config_io)

        assert full_config['ROOT']['command_path'] == [
            os.path.join(foo_dir, 'bar'),
            os.path.join(foo_dir, 'foobar'),
        ]
        assert full_config['ROOT']['foo'] == {
            'bar': 'barfoo',
            'one': 'two',
            'three': 4,
        }
        assert full_config['ROOT']['one'] == 1.1



import os
import sys
from argparse import ArgumentParser

from dodo_commands import CommandError, Dodo
from dodo_commands.framework.config_io import ConfigIO


def _args():
    parser = ArgumentParser(
        description="Enable or disable a layer in the configuration.")
    parser.add_argument('layer', nargs='?', default=False)
    parser.add_argument('value', nargs='?', default=False)
    parser.add_argument('--force', action="store_true")
    args = Dodo.parse_args(parser)
    return args


def _layer_name(x):
    try:
        return x.split(".")[0]
    except ValueError:
        return None


def _update_list_of_layers(layers, new_layer_name, new_layer_variant):
    def _new_layer():
        return "%s.%s.yaml" % (new_layer_name, new_layer_variant)

    result = []
    found = False
    for layer in layers:
        if _layer_name(layer) == new_layer_name:
            found = True
            result.append(_new_layer())
        else:
            result.append(layer)
    if not found:
        result.append(_new_layer())

    return result


def _layer_value(layers, layer):
    for l in layers:
        if l.startswith(layer + "."):
            try:
                parts = l.split(".")
                if len(parts) == 3:
                    return parts[1]
            except ValueError:
                pass
    return None


def _layers(config, filter_abs_paths=False):
    layers = config.get('LAYERS', [])
    if filter_abs_paths:
        layers = [
            x for x in layers
            if not os.path.exists(os.path.expanduser(os.path.dirname(x)))
        ]
    return layers


if Dodo.is_main(__name__, safe=False):
    args = _args()
    config = ConfigIO().load()

    if args.layer == False:  # noqa
        filtered_layers = _layers(config, filter_abs_paths=True)
        for layer in filtered_layers:
            name = _layer_name(layer)
            if name:
                value = _layer_value(filtered_layers, name)
                if value:
                    print(name + ': ' + _layer_value(filtered_layers, name))
        sys.exit(0)

    layers = _layers(config)
    if args.value == False:  # noqa
        print(_layer_value(layers, args.layer))
        sys.exit(0)

    layer_file = os.path.join(Dodo.get("/ROOT/config_dir"),
                              "%s.%s.yaml" % (args.layer, args.value))

    if not args.force and not os.path.exists(layer_file):
        raise CommandError("Layer file %s does not exist" % layer_file)

    config['LAYERS'] = _update_list_of_layers(layers, args.layer, args.value)
    ConfigIO().save(config)



"""Decorates command lines with docker arguments."""
from fnmatch import fnmatch

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.args_tree import ArgsTreeNode
from dodo_commands.framework.config import merge_into_config
from dodo_commands.framework.decorator_utils import uses_decorator

local = plumbum.local


def _is_tuple(x):
    return hasattr(x, "__len__") and not isinstance(x, type(str()))


class Decorator:  # noqa
    def is_used(self, config, command_name, decorator_name):
        return uses_decorator(config, command_name, decorator_name)

    @classmethod
    def merged_options(cls, get_config, command_name):
        merged = {}
        docker_options = get_config("/KUBERNETES", {})
        for patterns, docker_config in docker_options.items():
            should_merge = False
            for pattern in patterns if _is_tuple(patterns) else [patterns]:
                if pattern.startswith("!"):
                    should_merge = should_merge and not fnmatch(
                        command_name, pattern[1:]
                    )
                else:
                    should_merge = should_merge or fnmatch(command_name, pattern)
            if should_merge:
                merge_into_config(merged, docker_config)
        return merged

    @classmethod
    def kubectl_node(cls, get_config, command_name, cwd):
        merged = cls.merged_options(get_config, command_name)
        container_name = merged.get("container_name")
        kubectl = local["kubectl"]

        if container_name:
            full_container_name = kubectl(
                "get",
                "pods",
                "-l=app.kubernetes.io/name=" + container_name,
                "-o",
                'jsonpath="{.items[0].metadata.name}"',
            )

            return ArgsTreeNode(
                "kubectl",
                args=["kubectl", "exec", "-it", full_container_name[1:-1], "--"],
            )
        else:
            raise CommandError(
                "No kubernetes container name in /KUBERNETES for command %s"
                % command_name
            )

    def add_arguments(self, parser):  # noqa
        pass

    def modify_args(self, command_line_args, args_tree_root_node, cwd):  # override
        kubectl_node = self.kubectl_node(Dodo.get_config, Dodo.command_name, cwd)

        kubectl_node.add_child(args_tree_root_node)
        return kubectl_node, local.cwd



import sys

from dodo_commands.framework.config import expand_keys
from dodo_commands.framework.config_expander import get_key_expressions
from dodo_commands.framework.config_key import Key


class ConfigArg:
    def __init__(self, config_key, *args, **kwargs):
        self.config_key = config_key
        self.args = args
        self.kwargs = kwargs
        self.use_eval = bool(get_key_expressions(config_key))

    @property
    def arg_name(self):
        return self.args[0].strip("-").replace("-", "_")

    def _key(self, config):
        return Key(config, self.config_key)

    def exists(self, config):
        if self.use_eval:
            return self.get(config) is not None
        return self._key(config).exists()

    def get(self, config):
        if self.use_eval:
            return expand_keys(config, self.config_key)
        return self._key(config).get()



def add_config_args(parser, config, config_args):
    show_help = "--help" in sys.argv

    for config_arg in config_args or []:
        key_exists = config_arg.exists(config)

        if show_help or not key_exists:
            kwargs = dict(config_arg.kwargs)
            help_text = kwargs.get("help") or ""
            sep = ". " if help_text else ""
            if key_exists:
                value = str(config_arg.get(config))
                formatted_value = value[:50] + (value[50:] and "...")
                extra_help = "Read from config: %s = %s." % (
                    config_arg.config_key,
                    formatted_value,
                )
            else:
                extra_help = "Configuration key is %s" % config_arg.config_key

            kwargs["help"] = help_text + sep + extra_help
            parser.add_argument(*config_arg.args, **kwargs)



# -*- coding: utf-8 -*-
"""Utilities."""
import os
import sys

from dodo_commands.dependencies.get import plumbum, six

raw_input = six.moves.input


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = raw_input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


class classproperty(object):  # noqa
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        """Return read-only property value."""
        return self.f(owner)


def bordered(text):
    lines = text.splitlines()
    width = max(len(s) for s in lines)
    tl, tr, bl, br, v, h = u'┌', u'┐', u'└', u'┘', u'│', u'─'
    res = [tl + h * width + tr]
    for s in lines:
        res.append(v + (s + ' ' * width)[:width] + v)
    res.append(bl + h * width + br)
    return '\n'.join(res)


def is_using_system_dodo():
    import dodo_commands
    dodo_commands_path = dodo_commands.__path__[0]
    return os.path.realpath(dodo_commands_path) == dodo_commands_path


try:
    import textwrap
    textwrap.indent
except AttributeError:  # undefined function (wasn't added until Python 3.3)

    def indent(text, amount, ch=' '):
        padding = amount * ch
        return ''.join(padding + line for line in text.splitlines(True))
else:

    def indent(text, amount, ch=' '):
        return textwrap.indent(text, amount * ch)


def make_executable(script_filename):
    st = os.stat(script_filename)
    os.chmod(script_filename, st.st_mode | 0o111)


def symlink(source, link_name):
    import os
    os_symlink = getattr(os, "symlink", None)
    if callable(os_symlink):
        os_symlink(source, link_name)
    else:
        import ctypes
        csl = ctypes.windll.kernel32.CreateSymbolicLinkW
        csl.argtypes = (ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32)
        csl.restype = ctypes.c_ubyte
        flags = 1 if os.path.isdir(source) else 0
        if csl(link_name, source, flags) == 0:
            raise ctypes.WinError()


def chop(path):
    while path and path[-1] in ('\n', '\r'):
        path = path[:-1]
    return path


def _log(string):
    with open('/tmp/.dodo.log', 'w') as ofs:
        ofs.write(string + '\n')


class EnvironMemo:
    def __init__(self, extra_vars):
        self.extra_vars = extra_vars
        self.memo = {}

    def __enter__(self):  # noqa
        self.memo = os.environ.copy()
        os.environ.update(self.extra_vars)

    def __exit__(self, type, value, traceback):  # noqa
        os.environ.clear()
        os.environ.update(self.memo)


def exe_exists(exe):
    try:
        plumbum.local[exe]
        return True
    except plumbum.commands.processes.CommandNotFound:
        return False


class classproperty(property):  # noqa
    def __get__(self, cls, owner):  # noqa
        return classmethod(self.fget).__get__(None, owner)()


def maybe_list_to_list(maybe_list):
    return (maybe_list if isinstance(maybe_list, list) else [maybe_list])


def sh_cmd(cmd):
    return ['sh', '-c', cmd]


def xpath_to_string(xpath):
    return '/' + '/'.join([str(x) for x in xpath])



import os
import sys
from argparse import ArgumentParser

from dodo_commands import Dodo
from dodo_commands.dependencies import yaml_round_trip_load
from dodo_commands.dependencies.get import plumbum, semantic_version
from dodo_commands.framework import get_version
from dodo_commands.framework.config import Paths
from dodo_commands.framework.util import bordered

Version = semantic_version.Version
local = plumbum.local


def _args():
    parser = ArgumentParser(
        description="Compare configuration version to "
        "version in original project config file."
    )
    parser.add_argument(
        "--dodo",
        dest="check_dodo",
        action="store_true",
        help=(
            "Check that the installed dodo commands version satisfies the "
            "minimal dodo commands version in the dodo config "
            "(/ROOT/required_dodo_commands_version)"
        ),
    )
    parser.add_argument(
        "--config",
        dest="check_config",
        action="store_true",
        help=(
            "Check that the version field in the local dodo commands config "
            "(/ROOT/version) is up-to-date with the version in the shared "
            "dodo commands config."
        ),
    )
    args = Dodo.parse_args(parser)
    return args


def _shared_config_dir():
    src_dir = Dodo.get("/ROOT/src_dir", None)
    default_shared_config_dir = (
        os.path.join(src_dir, "extra", "dodo_commands", "res") if src_dir else None
    )
    return Dodo.get("/ROOT/shared_config_dir", default_shared_config_dir)


def _get_root_config_field(config_filename, field_name):
    with open(config_filename) as f:
        config = yaml_round_trip_load(f.read())
    return config.get("ROOT", {}).get(field_name, "")


def _partial_sem_version(version):
    v = Version(version)
    return Version("%s.%s" % (v.major, v.minor), partial=True)


def _config_filename():
    return os.path.join(Paths().config_dir(), "config.yaml")


def check_dodo_commands_version():  # noqa
    required_version = _get_root_config_field(
        _config_filename(), "required_dodo_commands_version"
    )
    if required_version:
        actual_version = get_version()
        if Version(actual_version) < Version(required_version):
            sys.stdout.write(
                bordered(
                    'The dodo_commands package needs to be upgraded (%s < %s). Tip: use "dodo upgrade"'
                    % (actual_version, required_version,),
                )
            )
            sys.stdout.write("\n")


def check_config_version():  # noqa
    if not _shared_config_dir():
        return

    original_file = os.path.join(_shared_config_dir(), "config.yaml")
    if not os.path.exists(original_file):
        return

    original_version = _get_root_config_field(original_file, "version")
    if not original_version:
        sys.stderr.write("No version found in original file %s\n" % original_file)
        return

    copied_file = os.path.join(Paths().config_dir(), "config.yaml")
    copied_version = _get_root_config_field(copied_file, "version")
    if not copied_version:
        sys.stderr.write("No version found in user managed file %s\n" % copied_file)
        return

    if _partial_sem_version(copied_version) < _partial_sem_version(original_version):
        sys.stdout.write(
            bordered(
                'Configuration needs update (%s < %s). Tip: use "dodo diff ."\n'
                % (copied_version, original_version,)
            )
        )
        sys.stdout.write("\n")


def check_config_changes():  # noqa
    if os.path.exists(os.path.join(Paths().config_dir(), ".git")):
        with local.cwd(Paths().config_dir()):
            changes = local["git"]("status", "--short")
            if changes:
                sys.stdout.write(
                    bordered(
                        'Configuration has changes (tip: use "dodo commit-config"):\n'
                        + changes
                    )
                )
                sys.stdout.write("\n")


if Dodo.is_main(__name__, safe=False):
    args = _args()
    if args.check_dodo:
        check_dodo_commands_version()
    if args.check_config:
        check_config_version()
        check_config_changes()



from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.config import build_config, extend_command_path
from dodo_commands.framework.config_layers import get_conflicts_in_layer_paths
from dodo_commands.framework.container.facets import Config, Layers, i_, o_, register


# CONFIG
@register(i_(Layers, "selected_layer_by_path"))
def check_conflicts_in_selected_layer_paths(selected_layer_by_path):
    conflicts = get_conflicts_in_layer_paths(selected_layer_by_path.keys())
    for path, conflicting_path in conflicts:
        raise CommandError("Conflicting layers: %s and %s" % (path, conflicting_path))
    return dict()


# CONFIG
@register(
    i_(Layers, "selected_layer_by_path"),
    i_(Layers, "metadata_by_layer_name"),
    o_(Config, "config"),
    o_(Config, "warnings"),
)
def build_from_selected_layers(
    selected_layer_by_path, metadata_by_layer_name,
):
    def get_selected_layers():
        return selected_layer_by_path.values()

    x = get_selected_layers()
    # [layer]
    config, warnings = build_config(x)
    config = extend_command_path(config)

    return dict(config=config, warnings=warnings)



import os
import sys
from argparse import ArgumentParser
from collections import defaultdict

from dodo_commands import Dodo
from dodo_commands.dependencies import yaml_round_trip_dump, yaml_round_trip_load
from dodo_commands.framework import ramda as R
from dodo_commands.framework.command_map import get_command_map
from dodo_commands.framework.command_path import get_command_dirs_from_config
from dodo_commands.framework.config import build_config
from dodo_commands.framework.config_layers import layer_filename_superset
from dodo_commands.framework.get_aliases import get_aliases
from dodo_commands.framework.version import get_version


def _args():
    parser = ArgumentParser(description="Show help on a command")
    parser.add_argument("command", nargs="*")
    parser.add_argument("--commands", action="store_true")
    parser.add_argument("--aliases", action="store_true")
    args = Dodo.parse_args(parser)
    return args


def _header_section():
    prog_name = os.path.basename(sys.argv[0])

    lines = [
        "",
        "Version %s (%s --version)." % (get_version(), prog_name),
        "To get help on a specific command, run it with the --help flag.",
        "Available commands (dodo help --commands):\n\n",
    ]
    return "\n".join(lines)


def _commands_section(command_map, inferred_commands, use_columns=True):
    def format_name(command_name):
        inferred_layer = inferred_commands.get(command_name, None)
        return (
            "%s (%s)" % (command_name, inferred_layer)
            if inferred_layer
            else command_name
        )

    """Return the script's main help text, as a string."""
    if not use_columns:
        lines = [x for x in sorted(command_map.keys())]
    else:
        lines = []

        max_name_size = -1
        command_groups = defaultdict(lambda: [])
        for command_name, command_map_item in sorted(command_map.items()):
            formatted_name = format_name(command_name)
            max_name_size = max(max_name_size, len(formatted_name))
            command_groups[command_map_item.group].append(formatted_name)

        groups = sorted(
            list(command_groups.values()), key=lambda x: len(x), reverse=True
        )
        for group in groups:
            group.append("")

        col_width = max_name_size + 10
        rows, cols = os.popen("stty size", "r").read().split()
        nr_cols = max(1, int(int(cols) / col_width))

        head, tail = groups[:nr_cols], groups[nr_cols:]
        while any([x is not None for x in head]):
            names = []
            for group_idx, group in enumerate(head):
                names.append("" if group is None else group.pop(0))
                if group is not None and len(group) == 0:
                    head[group_idx] = tail.pop(0) if tail else None
            lines.append("".join(name.ljust(col_width) for name in names))

    return "\n".join(lines)


def _collect_command_dirs(
    config,
    config_io,
    layer_names_by_command_dir,
    command_aliases,
    metadata_by_layer_name,
    layer_by_target_path,
):
    config_memo = yaml_round_trip_dump(config)

    for layer_name, layer_metadata in metadata_by_layer_name.items():

        def has_command_path(layer):
            return R.path_or(None, "ROOT", "command_path")(layer)

        layer_filenames = layer_filename_superset(
            [layer_metadata.target_path], config_io=config_io
        )
        extra_layers = R.map(config_io.load)(layer_filenames)
        if R.filter(has_command_path)(extra_layers):
            base_config = yaml_round_trip_load(config_memo)
            updated_config, warnings = build_config([base_config] + extra_layers)
            for command_dir in get_command_dirs_from_config(updated_config):
                layer_names = layer_names_by_command_dir[command_dir]
                _add_to_layer_names(layer_names, layer_name)

        layer = layer_by_target_path[layer_metadata.target_path]
        for command_alias in get_aliases(layer).items():
            alias_prefix = (
                ""
                if command_alias[0] in layer_metadata.inferred_commands
                else (layer_name + ".")
            )
            cmd_prefix = layer_name + "."
            command_aliases[alias_prefix + command_alias[0]] = (
                cmd_prefix + command_alias[1]
            )


def _print_command_dirs(args, layer_names_2_command_dirs, inferred_commands):
    if not args.commands:
        sys.stdout.write(_header_section())

    root_set = ("",)
    command_map = get_command_map(layer_names_2_command_dirs[root_set])
    sys.stdout.write(
        _commands_section(command_map, inferred_commands, use_columns=not args.commands)
        + "\n"
    )

    for layer_names, command_dirs in layer_names_2_command_dirs.items():
        if layer_names == root_set:
            continue

        sys.stdout.write("\nCommands for layer(s): %s\n\n" % ", ".join(layer_names))
        command_map = get_command_map(command_dirs)
        sys.stdout.write(
            _commands_section(
                command_map, inferred_commands, use_columns=not args.commands
            )
            + "\n"
        )


def _print_layer_names(metadata_by_layer_name):
    layer_names = metadata_by_layer_name.keys()
    col_width = max(len(x) for x in layer_names)

    sys.stdout.write("Layers:\n\n")
    for layer_name, layer_metadata in sorted(metadata_by_layer_name.items()):
        sys.stdout.write(
            "%s = %s\n" % (layer_name.ljust(col_width), layer_metadata.target_path)
        )


def _print_command_aliases(command_aliases, inferred_commands):
    if command_aliases:
        col_width = max(len(x) for x in command_aliases.keys())

        sys.stdout.write("Command aliases:\n\n")
        for alias, target_path in sorted(command_aliases.items()):
            sys.stdout.write("%s = %s\n" % (alias.ljust(col_width), target_path))


def _add_to_layer_names(layer_names, layer_name):
    if "" in layer_names:
        return
    layer_names.add(layer_name)


if Dodo.is_main(__name__, safe=True):
    args = _args()

    if args.command:
        Dodo.run([sys.argv[0], args.command, "--help"])
    else:
        ctr = Dodo.get_container()

        command_dirs = ctr.commands.command_dirs

        layer_names_by_command_dir = defaultdict(lambda: set())
        for command_dir in command_dirs:
            layer_names = layer_names_by_command_dir[command_dir]
            _add_to_layer_names(layer_names, "")

        command_aliases = dict(ctr.commands.aliases)
        _collect_command_dirs(
            ctr.config.config,
            ctr.layers.config_io,
            layer_names_by_command_dir,
            command_aliases,
            ctr.layers.metadata_by_layer_name,
            ctr.layers.layer_by_target_path,
        )

        layer_names_2_command_dirs = defaultdict(lambda: list())
        for command_dir, layer_names in layer_names_by_command_dir.items():
            layer_names_2_command_dirs[tuple(layer_names)].append(command_dir)

        if args.aliases:
            _print_command_aliases(
                command_aliases, ctr.commands.layer_name_by_inferred_command
            )
            sys.stdout.write("\n")
        else:
            _print_command_dirs(
                args,
                layer_names_2_command_dirs,
                ctr.commands.layer_name_by_inferred_command,
            )

            if not args.commands:
                if ctr.layers.metadata_by_layer_name:
                    _print_layer_names(dict(ctr.layers.metadata_by_layer_name))



class LayerMetadata:
    def __init__(self, target_path, inferred_commands, group_name):
        self.target_path = target_path
        self.inferred_commands = inferred_commands
        self.group_name = group_name




import os

from dodo_commands import Dodo
from dodo_commands.framework.config import Paths


def _args():
    Dodo.parser.add_argument('--alt', help='Run an alternative git command')
    Dodo.parser.add_argument('--message',
                             '-m',
                             dest='message',
                             help='The commit message')
    args = Dodo.parse_args()
    args.cwd = Paths().config_dir()
    return args


if Dodo.is_main(__name__, safe=True):
    args = _args()

    if args.alt:
        Dodo.run(['git', *args.alt.split()], cwd=args.cwd)
    else:
        if not os.path.exists(os.path.join(args.cwd, '.git')):
            Dodo.run(['git', 'init'], cwd=args.cwd)

        Dodo.run(['git', 'add', '-A'], cwd=args.cwd)
        Dodo.run(
            ['git', 'commit', '-m', args.message or 'Update configuration'],
            cwd=args.cwd)



from dodo_commands.framework.command_map import get_command_map as _get_command_map
from dodo_commands.framework.command_path import get_command_dirs_from_config
from dodo_commands.framework.container.facets import (
    Commands,
    Config,
    Layers,
    i_,
    o_,
    register,
)
from dodo_commands.framework.get_aliases import get_aliases
from dodo_commands.framework.inferred_commands import (
    get_inferred_command_map as _get_inferred_command_map,
)


# COMMANDS
@register(
    i_(Commands, "global_aliases"),
    i_(Layers, "metadata_by_layer_name"),
    i_(Layers, "layer_by_target_path"),
    o_(Commands, "layer_name_by_inferred_command"),
)
def get_inferred_command_map(
    global_aliases, metadata_by_layer_name, layer_by_target_path
):
    return dict(
        layer_name_by_inferred_command=_get_inferred_command_map(
            global_aliases, metadata_by_layer_name, layer_by_target_path,
        )
    )


# COMMANDS
@register(
    i_(Config, "config"),
    #
    o_(Commands, "aliases_from_config"),
)
def get_aliases_from_config(config):
    return dict(aliases_from_config=get_aliases(config))


# COMMANDS
@register(
    i_(Config, "config"), o_(Commands, "command_dirs"), o_(Commands, "command_map")
)
def get_command_map(config):
    command_dirs = get_command_dirs_from_config(config)
    return dict(command_dirs=command_dirs, command_map=_get_command_map(command_dirs),)



from dodo_commands.framework.util import make_executable


def create_dodo_script(env, dodo_file, python_path):
    with open(dodo_file, "w") as ofs:

        def add(x):
            ofs.write((x + "\n").format(python_path=python_path, env=env))

        if python_path:
            add("#!{python_path}")
        else:
            add("#!/usr/bin/env python")

        add("import os")
        add("import sys")
        add("")
        add("from dodo_commands.framework import execute_from_command_line")
        add("")
        add("if __name__ == \"__main__\":")
        if env:
            add("    os.environ[\"DODO_COMMANDS_ENV\"] = \"{env}\"")
        add("    execute_from_command_line(sys.argv)")
    make_executable(dodo_file)



import os

from dodo_commands.dependencies import yaml_round_trip_dump
from dodo_commands.framework.paths import Paths


def create_config_dir(config_dir):
    """Install the dir with dodo_commands resources."""
    os.makedirs(config_dir)
    config_filename = os.path.join(config_dir, "config.yaml")
    default_config = {
        'ROOT': {
            'command_path': [
                os.path.join(Paths().default_commands_dir(expanduser=False),
                             '*')
            ],
            'version':
            '1.0.0'
        }
    }
    with open(config_filename, "w") as f:
        f.write(yaml_round_trip_dump(default_config))

    os.makedirs(os.path.join(config_dir, ".dodo-start-env"))



import os

from dodo_commands import Dodo


def _args():
    Dodo.parser.description = "Restart the docker service"
    args = Dodo.parse_args()
    return args


# Use safe=False if the script makes changes other than through Dodo.run
if Dodo.is_main(__name__, safe=True):
    args = _args()
    memo = os.environ.get('__DODO_UNIVERSAL_CONFIRM__')
    os.environ['__DODO_UNIVERSAL_CONFIRM__'] = "1"
    Dodo.run(["sudo", "service", "docker", "stop"])
    if memo is None:
        del os.environ['__DODO_UNIVERSAL_CONFIRM__']
    else:
        os.environ['__DODO_UNIVERSAL_CONFIRM__'] = memo
    Dodo.run(["sudo", "service", "docker", "restart"])



import os

from dodo_commands import CommandError, Dodo


def _args():
    Dodo.parser.add_argument(
        "name",
        choices=Dodo.get("/DOCKER_IMAGES").keys(),
        help="Key to look up in /DOCKER_IMAGES",
    )
    Dodo.parser.add_argument(
        "build_args", nargs="*", help="Extra args to pass to docker build"
    )
    args = Dodo.parse_args()
    args.build_dir = Dodo.get(
        "/DOCKER_IMAGES/{name}/build_dir".format(name=args.name), "."
    )
    args.docker_file = Dodo.get(
        "/DOCKER_IMAGES/{name}/docker_file".format(name=args.name), "Dockerfile"
    )
    args.extra_dirs = Dodo.get(
        "/DOCKER_IMAGES/{name}/extra_dirs".format(name=args.name), []
    )
    args.docker_image = Dodo.get("/DOCKER_IMAGES/{name}/image".format(name=args.name))
    return args


def _copy_extra_dirs(local_dir, extra_dirs):
    # todo: convert extra dirs config mapping into a list of key=val
    for extra_dir in extra_dirs or []:
        extra_dir_name, extra_dir_path = extra_dir.split("=")
        local_path = os.path.join(local_dir, extra_dir_name)
        if os.path.exists(local_path):
            raise CommandError("Cannot copy to existing path: " + local_path)
        if not os.path.exists(extra_dir_path):
            raise CommandError("Cannot copy from non-existing path: " + extra_dir_path)

        Dodo.run(["cp", "-rf", extra_dir_path, local_path])
        if os.path.abspath(local_dir).startswith(os.path.abspath(extra_dir_path)):
            rp = os.path.relpath(local_dir, extra_dir_path)
            dead_path = os.path.join(local_path, rp)
            if os.path.exists(dead_path):
                Dodo.run(["rm", "-rf", dead_path])


def _remove_extra_dirs(local_dir, extra_dirs):
    for extra_dir in extra_dirs or []:
        extra_dir_name, extra_dir_path = extra_dir.split("=")
        local_path = os.path.join(local_dir, extra_dir_name)
        Dodo.run(["rm", "-rf", local_path])


if Dodo.is_main(__name__):
    args = _args()
    Dodo.safe = len(args.extra_dirs) == 0

    _copy_extra_dirs(args.build_dir, args.extra_dirs)

    try:
        Dodo.run(
            [
                "docker",
                "build",
                "-t",
                args.docker_image,
                "-f",
                args.docker_file,
                *args.build_args,
                ".",
            ],
            cwd=args.build_dir,
        )  # noqa
    finally:
        _remove_extra_dirs(args.build_dir, args.extra_dirs)



import os
import sys

from dodo_commands.framework.command_error import CommandError  # noqa
from dodo_commands.framework.command_map import execute_script
from dodo_commands.framework.singleton import Dodo  # noqa
from dodo_commands.framework.version import get_version


def _handle_exception(e):
    if '--traceback' in sys.argv or not isinstance(e, CommandError):
        raise
    sys.stderr.write('%s: %s\n' % (e.__class__.__name__, e))
    sys.exit(1)


def execute_from_command_line(argv):
    os.environ['__DODO__'] = '1'

    try:
        ctr = Dodo.get_container()

        # TODO: move to action_execute_command
        if ctr.command_line.command_name == '--version':
            print(get_version())
            sys.exit(0)

        if (ctr.command_line.is_trace):
            sys.stderr.write("%s\n" % ctr.command_line.get_trace())
            sys.exit(0)

        if ctr.command_line.command_name not in ctr.commands.command_map:
            raise CommandError("Unknown dodo command: %s" %
                               ctr.command_line.command_name)

        execute_script(ctr.commands.command_map, ctr.command_line.command_name)
    except KeyboardInterrupt:
        print('\n')
        sys.exit(1)
    except Exception as e:
        _handle_exception(e)



from dodo_commands import Dodo


def _args():
    Dodo.parser.description = "Run something on the commandline"

    Dodo.parser.add_argument("--cwd")
    Dodo.parser.add_argument("script_args", nargs="*")

    args = Dodo.parse_args()
    return args


# Use safe=False if the script makes changes other than through Dodo.run
if Dodo.is_main(__name__, safe=True):
    args = _args()
    Dodo.run(args.script_args, cwd=args.cwd or None)



import os
import sys
from contextlib import contextmanager

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.global_config import (global_config_get,
                                                   load_global_config_parser,
                                                   write_global_config_parser)
from dodo_commands.framework.paths import Paths

from ._create_env import forget_env, register_env, register_python_env

local = plumbum.local


def _args():
    Dodo.parser.add_argument("env", nargs="?")

    group = Dodo.parser.add_mutually_exclusive_group()
    group.add_argument("--create-python-env", action="store_true")
    group.add_argument("--use-python-env")

    group = Dodo.parser.add_mutually_exclusive_group()
    group.add_argument("--latest", action="store_true", dest="use_latest")
    group.add_argument("--create", action="store_true")
    group.add_argument("--init", action="store_true")
    group.add_argument("--forget", action="store_true")

    args = Dodo.parse_args()
    if args.create_python_env and not (args.create or args.init):
        raise CommandError(
            "The --create-python-env flag should be used together with"
            + " --init or --create"
        )
    return args


@contextmanager
def undo(undo_steps):
    try:
        yield
    except:  # noqa
        for undo_step in undo_steps:
            undo_step()
        raise


class Activator:
    def _report(self, x):
        sys.stderr.write(x)
        sys.stderr.flush()

    def init(self, env, use_latest):
        self.global_config = load_global_config_parser()
        if not self.global_config.has_section("recent"):
            self.global_config.add_section("recent")

        self.latest_env = global_config_get(self.global_config, "recent", "latest_env")
        self.shell = global_config_get(self.global_config, "settings", "shell", "bash")

        if env == "-":
            env = global_config_get(self.global_config, "recent", "previous_env")
            if not env:
                raise CommandError("There is no previous environment")

        if use_latest:
            if env:
                self._report("Options --latest and <env> are mutually exclusive\n")
                return False
            env = self.latest_env
            if not env:
                self._report("There is no latest environment\n")
                return False

        self.env = env
        self.env_dir = Paths().env_dir(self.env) if self.env else None

        return True

    def _config_dir(self, project_dir):
        return os.path.join(project_dir, ".dodo_commands")

    def _python_env_dir(self, create_python_env, use_python_env, project_dir):
        if use_python_env:
            return os.path.realpath(os.path.abspath(use_python_env))
        if create_python_env:
            return os.path.join(project_dir, ".env")
        return None

    def run(self, init, create, create_python_env, use_python_env, forget):
        if create:
            project_dir = os.path.expanduser(
                os.path.join(
                    self.global_config.get("settings", "projects_dir"), self.env
                )
            )
        elif init or create_python_env:
            project_dir = os.path.abspath(local.cwd)
        elif forget:
            pass
        else:
            if not os.path.exists(self.env_dir):
                raise CommandError(
                    "Environment not found: %s. Use the --create flag to create it\n"
                    % self.env
                )

        undo_steps = []
        with undo(undo_steps):
            if create or init:
                if os.path.exists(self.env_dir):
                    self._report(
                        "Cannot create an environment because the "
                        + "directory already exists: %s\n" % self.env_dir
                    )
                    return False

                self._report("Creating project directory %s ..." % project_dir)
                config_dir = self._config_dir(project_dir)
                register_env(
                    self.env,
                    project_dir,
                    config_dir,
                    self._python_env_dir(
                        create_python_env, use_python_env, project_dir
                    ),
                    undo_steps,
                )

            if create_python_env:
                register_python_env(
                    self.global_config.get("settings", "python_interpreter"),
                    self.env,
                    self._python_env_dir(
                        create_python_env, use_python_env, project_dir
                    ),
                    undo_steps,
                )

            if forget:
                forget_env(self.env)
                self.env = "default"

            if create or init or create_python_env:
                self._report(" done\n")

        if self.env != self.latest_env:
            self.global_config.set("recent", "previous_env", self.latest_env)
            self.global_config.set("recent", "latest_env", self.env)
            write_global_config_parser(self.global_config)

        activate_filename = os.path.join(self.env_dir, "activate.%s" % self.shell)

        if not forget:
            sys.stdout.write("source %s\n" % activate_filename)


if Dodo.is_main(__name__, safe=False):
    args = _args()
    if not args.env and not args.use_latest:
        raise CommandError("No environment was specified")

    activator = Activator()
    if activator.init(args.env, args.use_latest):
        activator.run(
            args.init,
            args.create,
            args.create_python_env,
            args.use_python_env,
            args.forget,
        )



"""Tests for all standard commands."""
import os
import shutil
import sys

from dodo_commands.dependencies import (yaml_round_trip_dump,
                                        yaml_round_trip_load)

docker_on_yaml = """
DOCKER_OPTIONS:
  "*":
    image: dodo_tutorial:1604

DOCKER_IMAGES:
  base:
    image: dodo_tutorial:1604

ROOT:
  command_path:
  - ${/ROOT/project_dir}/src/part4/after/src/commands
"""


class TestConfigIO:  # noqa
    @classmethod
    def _load_config(cls, config_filename):
        with open(config_filename) as f:
            return yaml_round_trip_load(f.read())

    @classmethod
    def _write_config(cls, config, config_filename):
        with open(config_filename, "w") as f:
            f.write(yaml_round_trip_dump(config))

    @classmethod
    def _set_config_version(cls, config_filename, version):
        config = cls._load_config(config_filename)
        config["ROOT"]["version"] = version
        cls._write_config(config, config_filename)

    @classmethod
    def _set_required_dodo_version(cls, config_filename, version):
        config = cls._load_config(config_filename)
        config["ROOT"]["required_dodo_commands_version"] = version
        cls._write_config(config, config_filename)

    @classmethod
    def _clear_layers(cls, config_filename):
        config = cls._load_config(config_filename)
        config["LAYERS"] = []
        cls._write_config(config, config_filename)

    def test_commands(self):  # noqa
        from dodo_commands.dependencies.get import plumbum

        local = plumbum.local

        dodo_test_dir = os.path.expanduser("~/projects/dodo_test")
        dodo_bin_dir = os.path.expanduser("~/.dodo_commands/bin")
        dodo_test_env_dir = os.path.expanduser("~/.dodo_commands/envs/dodo_test")
        config_dir = os.path.join(dodo_test_dir, ".dodo_commands")
        config_filename = os.path.join(config_dir, "config.yaml")
        skip_install = False

        if not skip_install:
            if os.path.exists(dodo_test_dir):
                shutil.rmtree(dodo_test_dir)
            if os.path.exists(dodo_test_env_dir):
                shutil.rmtree(dodo_test_env_dir)

            from dodo_commands.dodo import main

            sys.argv = ["dodo", "env", "--create", "--create-python-env", "dodo_test"]
            main()

        dodo = local[os.path.join(dodo_bin_dir, "dodo-dodo_test")]
        if not skip_install:
            dodo(
                "bootstrap",
                "src",
                "extra/dodo_commands/config",
                "--force",
                "--git-url",
                "https://github.com/mnieber/dodo_commands_tutorial.git",
            )

            # create shared config file
            extra_dir = os.path.join(dodo_test_dir, "extra")
            os.makedirs(extra_dir)

            config = self._load_config(config_filename)
            self._write_config(config, os.path.join(extra_dir, "config.yaml"))
            config["ROOT"]["shared_config_dir"] = "${/ROOT/project_dir}/extra"
            self._write_config(config, config_filename)

        # dodo which
        assert "dodo_test" == dodo("which")[:-1]
        assert dodo_test_dir == dodo("which", "--project-dir")[:-1]
        assert config_dir == dodo("which", "--config-dir")[:-1]
        assert config_filename == dodo("which", "--config")[:-1]
        assert "confirm, debugger, docker, pause" == dodo("which", "--decorators")[:-1]

        # dodo cd
        assert "cd " + dodo_test_dir == dodo("which --cd --project-dir")[:-1]

        # dodo check-version --config
        self._set_config_version(config_filename, "0.1.0")
        assert (
            "Configuration needs update (0.1.0 < 1.0.0)"
            in dodo("check-version", "--config")[:-1]
        )
        self._set_config_version(config_filename, "1.0.0")

        # dodo check-version --dodo
        self._set_required_dodo_version(config_filename, "10000.0.0")
        assert (
            "The dodo_commands package needs to be upgraded"
            in dodo("check-version", "--dodo")[:-1]
        )
        self._set_required_dodo_version(config_filename, dodo("--version")[:-1])
        assert "" == dodo("check-version", "--dodo")[:-1]

        # dodo diff
        result = dodo("diff", ".", "--echo").replace("\n", "")
        diff_tool = result.split()[0]
        assert "%s %s/extra %s/." % (diff_tool, dodo_test_dir, config_dir) == result

        # dodo docker
        with open(os.path.join(config_dir, "docker.on.yaml"), "w") as ofs:
            ofs.write(docker_on_yaml)

        dodo("layer", "docker", "on")
        result = (
            dodo("docker", "*", "--name", "foo", "--echo")
            .replace("\n", "")
            .replace("  \\", "")
            .replace("  ", " ")
        )
        assert (
            "docker run --name=foo --interactive --tty --rm --workdir=/ dodo_tutorial:1604 sh"
            == result
        )

        # dodo docker-build
        result = (
            dodo("docker-build", "--echo", "base").replace("\n", "").replace("  \\", "")
        )
        assert "docker build -t dodo_tutorial:1604 -f Dockerfile ." == result

        # dodo print-config
        assert "image: dodo_tutorial:1604" in dodo("print-config")

        # dodo drop-in
        tutorial_commands_dir = os.path.join(
            dodo_test_dir, "src/part4/after/src/commands"
        )
        drop_in_dir = os.path.join(tutorial_commands_dir, "drop-in")
        os.mkdir(drop_in_dir)
        with open(os.path.join(drop_in_dir, "bar.txt"), "w") as ofs:
            ofs.write("bar")
        target_bar_path = os.path.join(config_dir, "drops", "commands", "bar.txt")
        assert not os.path.exists(target_bar_path)
        dodo("drop-in", "commands")
        assert os.path.exists(target_bar_path)



from dodo_commands.dependencies.get import funcy


def filter(f):
    return funcy.partial(funcy.filter, f)


def remove_if(f):
    return filter(complement(f))


def split(pos):
    def f(s):
        return s[:pos], s[pos:]

    return f


def map(f):
    return funcy.partial(funcy.map, f)


def for_each(op):
    def f(seq):
        for x in seq:
            op(x)
        return seq

    return f


def pipe(*functions):
    return lambda x: funcy.rcompose(*functions)(x)


def ds(f):
    return lambda x: f(*x)


def debug(x):
    print(x)
    __import__("pud" + "b").set_trace()
    return x


def debug_as(f):
    return lambda x: debug(f(x))


def path(*keys):
    def f(x):
        result = x
        for key in keys:
            if not result or key not in result:
                raise Exception("Key not found %s" % key)
            result = result[key]
        return result

    return f


def path_or(default, *keys):
    def f(x):
        result = x
        for key in keys:
            if not result or key not in result:
                return default
            result = result[key]
        return result

    return f


def when(pred, transform):
    def f(x):
        return transform(x) if pred(x) else x

    return f


def prepend(x):
    def f(collection):
        return [x, *collection]

    return f


def equals(x):
    def f(y):
        return y == x

    return f


def split_when(pred, collection):
    return funcy.split_by(complement(pred), collection)


complement = funcy.complement
concat = funcy.concat
cut_prefix = funcy.cut_prefix
cut_suffix = funcy.cut_suffix
uniq = funcy.distinct
flatten = funcy.flatten
merge = funcy.merge
always = funcy.constantly
tap = funcy.tap



import re
from argparse import ArgumentParser

from dodo_commands.dependencies import yaml_round_trip_dump
from dodo_commands.framework.singleton import Dodo


def _args():
    parser = ArgumentParser(description="Print the full configuration.")
    parser.add_argument('key', nargs='?')
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__):
    args = _args()
    if args.key:
        prefix = "" if args.key.startswith("/") else "/"
        contents = Dodo.get(prefix + args.key)
        if isinstance(contents, str):
            print(contents)
        else:
            print("%s" % yaml_round_trip_dump(contents))
    else:
        content = re.sub(r'^([0-9_A-Z]+\:)$',
                         r'\n\1',
                         yaml_round_trip_dump(Dodo.get()),
                         flags=re.MULTILINE)
        print(re.sub(r'^\n\n', r'\n', content, flags=re.MULTILINE))



from collections import OrderedDict

from dodo_commands.framework import ramda as R
from dodo_commands.framework.config_layers import layer_filename_superset
from dodo_commands.framework.container.container import Container
from dodo_commands.framework.container.facets import (
    CommandLine,
    Layers,
    i_,
    o_,
    register,
)
from dodo_commands.framework.get_metadata_by_layer_name import (
    get_metadata_by_layer_name as _get_metadata_by_layer_name,
)
from dodo_commands.framework.load_named_layers import (
    load_named_layers as _load_named_layers,
)


# LAYERS
@register(
    i_(Container, "paths"),
    i_(Layers, "config_io"),
    i_(Layers, "root_layer_path"),
    #
    o_(Layers, "root_layer"),
)
def load_root_layer(paths, config_io, root_layer_path):
    return dict(
        root_layer=config_io.load(root_layer_path) if paths.project_dir() else {}
    )


# LAYERS
@register(
    i_(Layers, "root_layer"),
    #
    o_(Layers, "metadata_by_layer_name"),
)
def get_metadata_by_layer_name(root_layer):
    return dict(metadata_by_layer_name=_get_metadata_by_layer_name(root_layer))


# LAYERS
@register(
    i_(Layers, "config_io"),
    i_(Layers, "metadata_by_layer_name"),
    o_(Layers, "layer_by_target_path"),
)
def load_named_layers(config_io, metadata_by_layer_name):
    return dict(
        layer_by_target_path=_load_named_layers(config_io, metadata_by_layer_name,)
    )


# LAYERS
@register(
    i_(Layers, "config_io"),
    i_(Layers, "root_layer_path"),
    i_(Layers, "root_layer"),
    i_(CommandLine, "layer_paths", alt_name="command_line_layer_paths"),
    o_(Layers, "selected_layer_by_path"),
)
def select_layers(
    config_io, root_layer_path, root_layer, command_line_layer_paths,
):
    all_layer_paths = layer_filename_superset(
        R.uniq([root_layer_path] + command_line_layer_paths), config_io=config_io
    )

    selected_layer_by_path = OrderedDict()
    for layer_path in all_layer_paths:
        selected_layer_by_path[layer_path] = config_io.load(layer_path)
    return dict(selected_layer_by_path=selected_layer_by_path)



import os

from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.paths import Paths
from dodo_commands.framework.util import symlink


def env_var_template(env):
    return "export DODO_COMMANDS_ENV={env}\n".format(env=env)


def deactivate_template_bash():
    return """if [ -z "$VIRTUAL_ENV" ]
else
    echo "Deactivating virtual env: $VIRTUAL_ENV"
    deactivate
fi
"""


def deactivate_template_fish():
    return """if test -n "$VIRTUAL_ENV"
    echo "Deactivating virtual env: $VIRTUAL_ENV"
    deactivate
end
"""

def source_startup_files_fish(config_dir):
    return """for file in {config_dir}/.dodo-start-env/*
    source $file
end
""".format(config_dir=config_dir)

def activate_template(envs_dir, activate):
    return """source {envs_dir}/$DODO_COMMANDS_ENV/python_env_dir/bin/{activate}
""".format(envs_dir=envs_dir, activate=activate)


def create_env_dir(env, env_dir, project_dir, config_dir, python_env_dir):
    if os.path.exists(env_dir):
        raise CommandError("Environment dir already exists: %s" % env_dir)

    os.makedirs(env_dir)
    symlink(project_dir, os.path.join(env_dir, "project_dir"))
    symlink(config_dir, os.path.join(env_dir, "config_dir"))

    if python_env_dir:
        symlink(python_env_dir, os.path.join(env_dir, "python_env_dir"))

    with open(os.path.join(env_dir, "activate.bash"), "w") as ofs:
        ofs.write(env_var_template(env))
        ofs.write(deactivate_template_bash())
        if python_env_dir:
            ofs.write(activate_template(Paths().envs_dir(), 'activate'))

    with open(os.path.join(env_dir, "activate.fish"), "w") as ofs:
        ofs.write(env_var_template(env))
        ofs.write(deactivate_template_fish())
        ofs.write(source_startup_files_fish(config_dir))
        if python_env_dir:
            ofs.write(activate_template(Paths().envs_dir(), 'activate.fish'))
